/*
 
 Copyright (C) 2010 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */


#include <ctype.h>
#include "interface.h"
#include "seed.h"
#include "nuc_code.h"
#include "scoreboard.h"
#include "core.h"
#include "delve.h"
#include "align.h"
#include "dbio.h"

int seed_controller_thread(struct parameters* param)
{
	struct qs_struct* qs = 0;
	//unsigned long int fpos = 0u;
	int t;
	int interval;
	FILE *fastafile = 0; 
	qs = init_query_structure(qs,param);
	
	/*if(!isatty(0)){
		fastafile = stdin;
	}else{*/
		if (!(fastafile = fopen(param->infile[1], "r" ))){
			fprintf(stderr,"Cannot open query file '%s'\n", param->infile[1]);
			exit(-1);
		}
	//}
	
	qs->hash_t = malloc(sizeof(int*) * param->num_threads);
	qs->hash_index_t = malloc(sizeof(int*) * param->num_threads);
	
	qs->size = param->num_query;
	//while(fpos != FILE_END){
		//fpos = read_fasta_fastq(fastafile,qs,param,fpos);
	while(qs->size ==  param->num_query ){	
		qs = read_fasta_fastq2(fastafile,qs,param);
		fprintf(stderr,"%d read\n",qs->size);
		//qs = make_B(qs,qs->size);
		interval =  (int)((double)qs->size /(double)param->num_threads);
		
		for(t = 0;t < param->num_threads - 1;t++) {
			qs = make_hash_thread(qs,param->r,param->step,t, t*interval,t*interval+interval);

		}
		qs = make_hash_thread(qs,param->r,param->step,param->num_threads - 1, t*interval,qs->size);
		
		qs = search_fasta_fastq_thread(qs,param);
		
		for(t = 0;t < param->num_threads ;t++){
			free(qs->hash_t[t]);
			free(qs->hash_index_t[t]);
		}
		
	}
	rewind(fastafile);
	
	fclose(fastafile);
	free(qs->hash_t);
	free(qs->hash_index_t);
	free_query_structure(qs,param);
	return 1;
}

struct qs_struct* make_B(struct qs_struct* qs,int n)
{
	int i,j;
	struct seq_info* si = 0;
	for(i = 0; i < n;i++){
		si = qs->seq_info[i];
		for(j = 0; j < 4;j++){
		//	si->rB[j]= 0ul;
		//	si->fB[j]= 0ul;
		}
			
		for(j = 0; j < si->len;j++){
		//	si->fB[(int)(si->seq[j] &0x3u)] |= (1ul << (unsigned long int)(si->len-1-j));
		//	si->rB[(int)(si->reverse_seq[j] &0x3u)] |= (1ul << (unsigned long int)(si->len-1-j));
		}
	}
	return qs;
}


struct qs_struct* make_hash_thread(struct qs_struct* qs,int r,int step,int thread,int start,int stop)
{
	unsigned int key = 0;
	int i,j,c;
	int elements = 0;
	unsigned char* tmp = 0;
	int m;
	//int n = qs->size;
	fprintf(stderr,"thread:%d\n",thread);
	qs->hash_index_t[thread] = malloc(sizeof(int) * ((1 << r * 2) +1));
	
	for(i = 0; i < (1 << r * 2)+1;i++){
		qs->hash_index_t[thread][i] = 0;
	}
	
	for(i = start; i < stop;i++){
		m = qs->seq_info[i]->len;
		if(m > 64){
			m = 64;
		}
		tmp = qs->seq_info[i]->seq;//   sequences[i];
		for(j = 0;j + r  <= m;j+= step){
			key = 0;
			for(c = 0;c < r;c++){
				key = (key << 2u) | ((int)tmp[c]   & 0x3u);			
			}
			qs->hash_index_t[thread][key+1]++;
			tmp += step;
		}	
		//m = qs->seq_info[i]->len;
		tmp = qs->seq_info[i]->reverse_seq;//   sequences[i];
		j = r;
		
		while(j+step <= m){
			j += step;
		}
		
		tmp += m - j;
		
		
		//tmp += (m % r) % step;
		for(j = 0;j + r  <= m;j+=step){
			key = 0;
			for(c = 0;c < r;c++){
				key = (key << 2u) | ((int)tmp[c]& 0x3u);		
			}
			qs->hash_index_t[thread][key+1]++;
			tmp += step;
		}	
	}
	
	for(i = 1; i < (1 << r * 2)+1;i++){
		qs->hash_index_t[thread][i] +=  qs->hash_index_t[thread][i-1];
	}
	elements = qs->hash_index_t[thread][(1 << r * 2) ];
	
	qs->hash_t[thread]  = malloc(sizeof(int) * (elements));
	for(i = 0; i < elements;i++){
		qs->hash_t[thread][i] = -1;
	}
	for(i = start; i < stop;i++){
		m = qs->seq_info[i]->len;
		if(m > 64){
			m = 64;
		}
		tmp = qs->seq_info[i]->seq;//   sequences[i];
		for(j = 0;j + r  <= m;j+= step){
			key = 0;
			for(c = 0;c < r;c++){
				key = (key << 2u) | ((int)tmp[c]   & 0x3u);
				
			}
			
			for(c = qs->hash_index_t[thread][key];c < qs->hash_index_t[thread][key+1u];c++){
				if(qs->hash_t[thread] [c] == -1){
					qs->hash_t[thread] [c] = (i << 8u) | j;
					//qs->hash[c] = i;
					break;
				}
			}
			tmp += step;
		}	
		//m = qs->seq_info[i]->len;
		tmp = qs->seq_info[i]->reverse_seq;//   sequences[i];
		
		//for(j = 0; j < m;j++){
		//	fprintf(stderr,"%c",tmp[j]+65);
		//}
		//fprintf(stderr,"\n");
		
		j = r;
		
		while(j+step <= m){
			//fprintf(stderr,"%d	%d\n",j,m);
			j += step;
		}
		//elements = m - j;
		//fprintf(stderr,"offset:%d	%d	%d\n",elements,j,m);
		
		tmp += m - j;
		
		//tmp += (m % r) % step;
		
		for(j = 0;j + r  <= m;j+= step){
			key = 0;
			//for(c = 0;c < elements;c++){
			//	fprintf(stderr,"-");
			//}
			//for(c = 0;c < j;c++){
			//	fprintf(stderr,"-");
			//}
			for(c = 0;c < r;c++){
				key = (key << 2u) | ((int)tmp[c]& 0x3u);
			//	fprintf(stderr,"%c",tmp[c]+65);
			}
			//fprintf(stderr,"\n");
			for(c = qs->hash_index_t[thread][key];c < qs->hash_index_t[thread][key+1u];c++){
				if(qs->hash_t[thread][c] == -1){
					qs->hash_t[thread][c] = (i << 8u) | j | 0x80u;
					break;
				}
			}
			tmp += step;
		}	
		//fprintf(stderr,"\n");
	}
	//exit(0);
	return qs;
}


void* search_hash_thread(void *threadarg)
{
	
	struct seed_thread_data *data;
	data = (struct seed_thread_data *) threadarg;
	struct qs_struct* qs = data->qs;
	struct parameters* param = data->param;
	
	
	unsigned char* t = data->chromsome;
	long int n = data->len;
	long int r = (long int) param->r;
	long int offset = data->total_len;
	
	int thread = data->thread;
	
	int j;//,g;
	int i;
	unsigned char* target = 0;
	struct seq_info* si = 0;
	unsigned int key = 0;
	unsigned int mask = 0;
	int t_len = 0;
	long int seed_offset = 0;
	int strand = 0;
	unsigned long int new = 0;
	long int add = 0;
	
	int* index = qs->hash_index_t[thread];
	
	for(i = 0; i <  r;i++){
		mask = (mask << 2u) | 3u;
	}
	key = 0u;
	for(i = 0; i <  r-1;i++){
		key = (key << 2u) | (t[i] & 0x3u);
	}
	for(i = r-1; i < n;i++){
		key = ((key << 2u) | (t[i]  & 0x3u)) & mask;
		if(key){
			//g = 0;
			for(j = index[key];j < index[key+1u];j++){
				seed_offset = qs->hash_t[thread][j] & 0x7fu;
				
				si = qs->seq_info[ qs->hash_t[thread][j] >> 8u];
				//id = qs->hash[j] >> 8u;
				strand = qs->hash_t[thread][j] & 0x80u;
				add = ( (long int)i + ((long int) si->len - seed_offset)  -r + 10l); /// should be 10 ??? 
				//if( add > 0){
					target = t +add;  
				//}else{
				//	target = t; /// WRONG????? 
				//}
				
				//if( add + si->len < n){
				t_len = si->len + 10;
				//if(add - t_len < 0){
				//	t_len = add;
				//}
				//}else{
				//	t_len = n - add;
 				//}
				if( add <= n && add > ((long int) (si->len + 20))){
					if(!strand){
						if(si->last_f_test <= i){
							//fprintf(stderr,"ADD:%ld	LEN:%d\n",add,t_len);
							new = validate_bpm(target ,si->seq , t_len , si->len, offset+ add);
							
							//new = validate_bpm_pre(target ,si->fB , t_len , si->len, offset+ add);
							
							BinaryInsertionSortOne(si->match_units,LIST_STORE_SIZE ,new);
							si->last_f_test = i +  si->len  ;
						}
					}else{
						if(si->last_r_test <= i){
							new = validate_bpm(target ,si->reverse_seq , t_len , si->len,offset + add);
							//new = validate_bpm_pre(target ,si->rB , t_len , si->len, offset+ add);
							new |= 1ul ;
							BinaryInsertionSortOne(si->match_units,LIST_STORE_SIZE,new);
							si->last_r_test = i +   si->len ;
						}
					}
				}
			}
		}
		//if(i % 100000 == 0 && thread == 0){
		//	fprintf(stderr,"\r%8.2f percent done",(double)i /(double)n * 100);
		//}
	}
	
	//fprintf(stderr,"\r%8.2f percent done\n",100.0);
	pthread_exit((void *) 0);
}	




struct qs_struct* search_fasta_fastq_thread(struct qs_struct* qs,struct parameters* param) 
{
	static int header = 1;
	struct seed_thread_data* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct seed_thread_data)* param->num_threads);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	
	FILE *file;
	unsigned int string_pos = 0;
	int park_pos = -1;
	char line[MAX_LINE];
	int i,j,c,n;
	int seq_p = 0;
	int set = 0;
	long int len = 0;
	long int total_len = 0;
	unsigned char* chromsome =  malloc(sizeof(unsigned char)* 300000000);
	
	long int* lengths = 0; 
	lengths = malloc(sizeof(long int) * 50);
	unsigned char** chr_names = 0;
	chr_names = malloc(sizeof(unsigned char*) * 50);
	
	
	if (!(file = fopen(param->infile[2], "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->infile[2]);
		exit(-1);
	}
	c= -1;
	
	while(fgets(line, MAX_LINE, file)){
		//reading name .... 
		if((line[0] == '@' && !set)|| (line[0] == '>' && !set)){
			//set sequence length of previous read
			
			if(park_pos != -1){ // to avoid setting something before the first read was read
				
				
				interval =  (int)((double)qs->size /(double)param->num_threads);
				
				for(t = 0;t < param->num_threads - 1;t++) {
					thread_data_array[t].qs = qs;
					thread_data_array[t].param = param;
					thread_data_array[t].start = t*interval;
					thread_data_array[t].end = t*interval + interval;
					thread_data_array[t].thread = t;
					thread_data_array[t].chromsome = chromsome;
					thread_data_array[t].total_len = total_len;
					thread_data_array[t].len = len;
				}
				thread_data_array[param->num_threads - 1].qs = qs;
				thread_data_array[param->num_threads - 1].param = param;
				thread_data_array[param->num_threads - 1].start = t*interval;
				thread_data_array[param->num_threads - 1].end = qs->size;
				thread_data_array[param->num_threads - 1].thread = param->num_threads - 1;
				thread_data_array[param->num_threads - 1].chromsome = chromsome;
				thread_data_array[param->num_threads - 1].total_len = total_len;
				thread_data_array[param->num_threads - 1].len = len;
				
				
				
				
				rc = pthread_attr_init(&attr);
				if(rc){
					fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
					exit(-1);
				}
				pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
				
				for(t = 0;t < param->num_threads;t++) {
					rc = pthread_create(&threads[t], &attr, search_hash_thread, (void *) &thread_data_array[t]);
					if (rc) {
						fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
						exit(-1);
					}
				}
				
				pthread_attr_destroy(&attr);
				
				for (t = 0;t < param->num_threads;t++){
					rc = pthread_join(threads[t], NULL);
					if (rc){
						fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
						exit(-1);
					}
				}
				
				
				
				
				//qs =  search_hash(qs,chromsome,len,param->r , total_len);
				
				//exit(0);
				
				for(i = 0; i< qs->size;i++){
					qs->seq_info[i]->last_f_test = -1000;
					qs->seq_info[i]->last_r_test = -1000;
				}
				total_len += len;
				//fprintf(stderr,"	%d	%ud	%ud	(%f)\n",len,mers,total_len,(float)mers/(float)total_len*100.0f);
			}
			fprintf(stderr,"%s	%ld\n",line,total_len);
			//fprintf(stderr,"%s",line);
			
			n = strlen(line);
			c++;
			
			chr_names[c] = malloc(sizeof(unsigned char) * (n+1));
			for(i = 1; i < n-1;i++){
				chr_names[c][i-1] = line[i];
			}
			chr_names[c][n-2] = 0; 
			lengths[c] = len;
			
			string_pos = 0;			
			len = 0;
			seq_p = 1;
			set = 1;			
			park_pos++;
			//get ready to read quality if present  
		}else if(line[0] == '+' && !set){
			seq_p = 0;
			set = 1;
			//reading sequence or quality  
		}else{			
			if(!set){
				string_pos--;
			}
			
			if(seq_p){
				for(i = 0;i < MAX_LINE;i++){
					if(iscntrl((int)line[i])){
						chromsome[string_pos] = 0;
						//qt->string[string_pos] = 0;
						string_pos++;
						break;
					}
					
					len++;
					chromsome[string_pos] = nuc_code5[(int)line[i]];
					string_pos++;
				}
			}
			set = 0;
			
		}
	}
		
	interval =  (int)((double)qs->size /(double)param->num_threads);
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].qs = qs;
		thread_data_array[t].param = param;
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].thread = t;
		thread_data_array[t].chromsome = chromsome;
		thread_data_array[t].total_len = total_len;
		thread_data_array[t].len = len;
	}
	thread_data_array[param->num_threads - 1].qs = qs;
	thread_data_array[param->num_threads - 1].param = param;
	thread_data_array[param->num_threads - 1].start = t*interval;
	thread_data_array[param->num_threads - 1].end = qs->size;
	thread_data_array[param->num_threads - 1].thread = param->num_threads - 1;
	thread_data_array[param->num_threads - 1].chromsome = chromsome;
	thread_data_array[param->num_threads - 1].total_len = total_len;
	thread_data_array[param->num_threads - 1].len = len;
	
	
	
	
	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, search_hash_thread, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}
	c++;
	
	lengths[c] = len;
	chr_names[c] = 0;
	if(header){
		fprintf(outfile, "@HD\tVN:1.0\n");
		
		for(i = 0; i < c ;i++){
			fprintf(outfile, "@SQ\tSN:%s\tLN:%u",chr_names[i], (unsigned int) lengths[i+1]);
			if(param->assembly_version){
				fprintf(outfile, "\tAS:%s\n",param->assembly_version);
			}else{
				fprintf(outfile,"\n");
			}
		}
		fprintf(outfile, "@PG\tID:Delve\tVN:%s\tCL:%s\n",VERSION,param->command_line) ;
		header = 0;
	}
	
	
	
	for(i = 1; i <= c;i++){
		lengths[i] +=  lengths[i-1];
	}
	
	
	//total_len +=  len;
	for(i = 0; i< qs->size;i++){
		print_sam(qs->seq_info[i], chr_names, lengths ,c);
	}
	
	//exit(0);
	for(i = 0; i< qs->size;i++){
		for(j = 0; j < LIST_STORE_SIZE;j++){
			qs->seq_info[i]->match_units[j] = 0ul;
		}
	}
	//string_pos = 0;
	qs->q_num = 0;
	
	for(i = 0; i < c;i++){
		free(chr_names[i]);
	}
	
	free(thread_data_array);
	free(chr_names);
	free(lengths);
	free(chromsome);
	fclose(file);
	return qs;
}





int seed_controller(struct parameters* param)
{
	struct qs_struct* qs = 0;
	unsigned long int fpos = 0u;
	//int c;
	FILE *fastafile = 0; 
	qs = init_query_structure(qs,param);
	
	
	if (!(fastafile = fopen(param->infile[1], "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->infile[2]);
		exit(-1);
	}
	//c = 0;
	while(fpos != FILE_END){
		fpos = read_fasta_fastq(fastafile,qs,param,fpos);
		/*fprintf(stderr,"Chumk:%d\n",c);
		 for(i = 0; i < qs->size;i++){
		 fprintf(stderr,"%s	%s	",qs->seq_info[i]->sn,qs->seq_info[i]->qual);
		 for(j = 0;j < qs->seq_info[i]->len;j++){
		 fprintf(stderr,"%c",qs->seq_info[i]->seq[j] + 65);
		 }
		 fprintf(stderr,"\t");
		 for(j = 0;j < qs->seq_info[i]->len;j++){
		 fprintf(stderr,"%c",qs->seq_info[i]->reverse_seq[j] + 65);
		 }
		 fprintf(stderr,"\n");
		 
		 }*/
		//c++;
		
		qs = make_hash(qs,param->r);
		qs = search_fasta_fastq(qs,param);
		free(qs->hash);
		free(qs->hash_index);
		
	}
	rewind(fastafile);
	
	fclose(fastafile);
	return 1;
}

struct qs_struct* init_query_structure(struct qs_struct * qs,struct parameters* param)
{
	int i,j;
	
	qs = malloc(sizeof(struct qs_struct));	
	qs->q_num = 0;
	qs->min_len = 100000000;
	qs->max_len = -1;
	qs->string_len = 0;
	qs->seq_info = malloc(sizeof(struct seq_info*) *param->num_query);
	//qs->string = malloc(sizeof(unsigned char)*param->num_query*255);
	for(i = 0; i < param->num_query ;i++){
		qs->seq_info[i] = malloc(sizeof(struct seq_info));
		qs->seq_info[i]->sn = 0;
		qs->seq_info[i]->reverse_seq = 0;
		qs->seq_info[i]->seq = 0;
		qs->seq_info[i]->qual = 0;
		qs->seq_info[i]->last_f_test = -1000;
		qs->seq_info[i]->last_r_test = -1000;
		qs->seq_info[i]->match_units= malloc(sizeof(unsigned long int) *  LIST_STORE_SIZE);//param->reported_seed_hits ); 
		for(j = 0; j < LIST_STORE_SIZE;j++){
			qs->seq_info[i]->match_units[j] = 0ul;
		}
	}
	return qs;
}

void free_query_structure(struct qs_struct * qs,struct parameters* param)
{
	int i;
	
	for(i = 0; i < param->num_query ;i++){
		if(qs->seq_info[i]->sn){
			free(qs->seq_info[i]->sn);//
			free(qs->seq_info[i]->seq);
			free(qs->seq_info[i]->reverse_seq);
		}
		if(qs->seq_info[i]->qual){
			free(qs->seq_info[i]->qual);
		}
		free(qs->seq_info[i]->match_units); 
		free(qs->seq_info[i]);
		
	}

	free(qs->seq_info);  
	//free(qs->string); 
	
	free(qs);
}


struct qs_struct* read_fasta_fastq2(FILE *file, struct qs_struct* qs,struct parameters* param) 
{
	int park_pos = -1;
	char line[MAX_LINE];
	unsigned char rc[MAX_LINE];
	int i,j;
	int seq_p = 0;
	int set = 0;
	int len = 0;
	int fastq = 0;
	qs->size = 0;
	for(i = 0; i < param->num_query ;i++){
		//qs->seq_info[i] = malloc(sizeof(struct seq_info));
		if(qs->seq_info[i]->sn){
			free(qs->seq_info[i]->sn);//
			free(qs->seq_info[i]->seq);
			free(qs->seq_info[i]->reverse_seq);
		}
		if(qs->seq_info[i]->qual){
			//fprintf(stderr,"%s\n",qs->seq_info[i]->qual);
			free(qs->seq_info[i]->qual);
		}

		qs->seq_info[i]->sn = 0;
		qs->seq_info[i]->seq = 0;
		qs->seq_info[i]->reverse_seq = 0;
		qs->seq_info[i]->qual = 0;
		
		
		qs->seq_info[i]->last_f_test = -1000;
		qs->seq_info[i]->last_r_test = -1000;
		//qs->seq_info[i]->match_units= malloc(sizeof(unsigned long int) *  LIST_STORE_SIZE);//param->reported_seed_hits ); 
		for(j = 0; j < LIST_STORE_SIZE;j++){
			qs->seq_info[i]->match_units[j] = 0ul;
		}
	}
	while(fgets(line, MAX_LINE, file)){
		//reading name .... 
		//fprintf(stderr,"%s",line);
		//if(line[0] != '#'){ 
			if((line[0] == '@' && !set)|| (line[0] == '>' && !set)){
				//set sequence length of previous read
				//if(park_pos != -1){
				//	fprintf(stderr,"%s	%s	%s\n",ri[park_pos]->name,ri[park_pos]->seq);
				//}
				
				//check if there is still space.... 
				//if(param->num_query == size){
				//	fseek (file , -  strlen(line) , SEEK_CUR);
				
				//	return size;
				//}
				park_pos++;
				
				qs->seq_info[park_pos]->last_f_test = -1000;
				qs->seq_info[park_pos]->last_r_test = -1000;
				qs->seq_info[park_pos]->sn = 0;
				qs->seq_info[park_pos]->seq = 0;
				qs->seq_info[park_pos]->reverse_seq = 0;
				qs->seq_info[park_pos]->qual = 0;
				len = 0;
				seq_p = 1;
				for(i = 1;i < MAX_LINE;i++){
					len++;
					if(isspace((int)line[i])){
						break;
					}
					
				}
				qs->seq_info[park_pos]->sn = malloc(sizeof(unsigned char)* (len+1));
				for(i = 1;i < MAX_LINE;i++){
					
					if(isspace((int)line[i])){
						qs->seq_info[park_pos]->sn[i-1] = 0;
						break;
					}
					qs->seq_info[park_pos]->sn[i-1] = line[i];
				}
				//fprintf(stderr,"LEN:%d	%s\n",len,ri[park_pos]->name);
				
				set = 1;
				qs->size++;
				//get ready to read quality if present  
			}else if(line[0] == '+' && !set){
				seq_p = 0;
				set = 1;
				fastq = 1;
				//fprintf(stderr,"fastq	%d\n",fastq);
				//reading sequence or quality  
			}else{	
				if(set){
					if(seq_p){
						len = 0;
						for(i = 0;i < MAX_LINE;i++){
							
							if(isspace((int)line[i])){
								break;
							}
							len++;
							
						}
						//fprintf(stderr,"SEQ LEN:%d	%s\n",len,line);
						qs->seq_info[park_pos]->seq = malloc(sizeof(unsigned char)* (len+1));
						for(i = 0;i < MAX_LINE;i++){
							
							if(isspace((int)line[i])){
								qs->seq_info[park_pos]->seq[i] = 0;
								break;
							}
							qs->seq_info[park_pos]->seq[i] = nuc_code5[(int)line[i]];
						}
						qs->seq_info[park_pos]->len = len;
						
						
						reverse_complement(qs->seq_info[park_pos]->seq ,rc,len);
						qs->seq_info[park_pos]->reverse_seq = malloc(sizeof(unsigned char)* (len+1));
						for(i = 0; i < len;i++){
							qs->seq_info[park_pos]->reverse_seq[i] = rc[i];
						}
						qs->seq_info[park_pos]->reverse_seq[len] = 0;
						
						
						if(param->num_query == qs->size && fastq == 0){
							return qs;
						}
					}else{
						len = 0;
						for(i = 0;i < MAX_LINE;i++){
							len++;
							if(isspace((int)line[i])){
								break;
							}
							
						}
						//fprintf(stderr,"QUAL LEN:%d\n",len);
						qs->seq_info[park_pos]->qual = malloc(sizeof(unsigned char)* (len+1));
						for(i = 0;i < MAX_LINE;i++){
							
							if(isspace((int)line[i])){
								qs->seq_info[park_pos]->qual[i] = 0;
								break;
							}
							qs->seq_info[park_pos]->qual[i] = line[i];
						}
						
						if(param->num_query == qs->size){
							return qs;
						}
					}
				}
				set = 0;
			}
		//}
	}
	return qs;
}

unsigned long int read_fasta_fastq(FILE *file, struct qs_struct* qs,struct parameters* param,unsigned long int position) 
{
	
	unsigned int string_pos = 0;
	int park_pos = -1;
	char line[MAX_LINE];
	unsigned char rc[MAX_LINE];
	int i;//,j;
	int seq_p = 0;
	int set = 0;
	int len = 0;
	int size = 0;
	unsigned long int fpos = 0;
	
	fseek (file , position , SEEK_SET);
	//size = ftell (file);
	//fprintf(stderr,"Position = %d\n",size);
	
	while(fgets(line, MAX_LINE, file)){
		//reading name .... 
		//fprintf(stderr,"%s",line);
		
		if(line[0] != '#'){
		
		if((line[0] == '@' && !set)|| (line[0] == '>' && !set)){
			//set sequence length of previous read
			
			if(park_pos != -1){ // to avoid setting something before the first read was read
				//len = (int)strlen((char*)qs->seq_info[park_pos]->seq);
				qs->seq_info[park_pos]->len = len;
				
				if(len < qs->min_len){
					qs->min_len = len;
				}
				
				if(len > qs->max_len){
					qs->max_len = len;
				}			
				
				reverse_complement(qs->seq_info[park_pos]->seq ,rc,len);
				qs->seq_info[park_pos]->reverse_seq = qs->string + string_pos;
				for(i = 0; i < len;i++){
					qs->string[string_pos] = rc[i];
					string_pos++;
				}
				qs->string[string_pos] = 0;
				string_pos++;
				
				
			}
			
			//check if there is still space.... 
			if(string_pos + MAX_LINE >= (param->num_query *255) || size >= param->num_query){
				//fprintf(stderr,"\nLast_line = %s",line);
				//size = ftell (file);
				//fprintf(stderr,"Position = %d\n",size);
				fseek (file , -  strlen(line) , SEEK_CUR);
				fpos = ftell (file);
				qs->size = size;
				//fgets(line, MAX_LINE, file);
				//fprintf(stderr,"Last_line = %s",line);
				//fprintf(stderr,"Position = %d\n",fpos);
				return fpos;
				for(i= 0; i < park_pos;i++){
					fprintf(stderr,"%d	%s	%s	%s\n",qs->seq_info[i]->id_num,qs->seq_info[i]->sn,qs->seq_info[i]->seq,qs->seq_info[i]->qual);
				}
				size = 0;
				string_pos = 0;
				park_pos = -1;
			}
			len = 0;
			seq_p = 1;
			park_pos++;
			qs->q_num++;
			qs->seq_info[park_pos]->id_num = qs->q_num;
			qs->seq_info[park_pos]->sn =  qs->string + string_pos;
			qs->seq_info[park_pos]->qual = 0;
			qs->seq_info[park_pos]->seq = 0;
			qs->seq_info[park_pos]->last_f_test = -1000;
			qs->seq_info[park_pos]->last_r_test = -1000;
			for(i = 1;i < MAX_LINE;i++){
				if(iscntrl((int)line[i])){
					qs->string[string_pos] =0;
					string_pos++;
					break;
				}
				if(isspace((int)line[i])){
					qs->string[string_pos] = '_';
				}else{
					qs->string[string_pos] = line[i];
				}
				string_pos++;
			}
			set = 1;
			size++;
			//get ready to read quality if present  
		}else if(line[0] == '+' && !set){
			seq_p = 0;
			set = 1;
			//reading sequence or quality  
		}else{	
			if(set){
				if(seq_p){
					qs->seq_info[park_pos]->seq = qs->string + string_pos;
				}else{
					qs->seq_info[park_pos]->qual = qs->string + string_pos;
				}
			}else{
				string_pos--;
			}
			
			for(i = 0;i < MAX_LINE;i++){
				if(iscntrl((int)line[i])){
					qs->string[string_pos] = 0;
					string_pos++;
					break;
				}
				if(seq_p){
					//fprintf(stderr,"Reading seq:%c	%d\n",line[i],nuc_code5[line[i]]);
					len++;
					qs->string[string_pos] = nuc_code5[(int)line[i]];
				}else{
					qs->string[string_pos] = line[i];
				}
				string_pos++;
			}
			set = 0;
		}
		}
	}
	if(park_pos != -1){ // to avoid setting something before the first read was read
		//len = (int)strlen((char*)qs->seq_info[park_pos]->seq);
		qs->seq_info[park_pos]->len = len;
		//qs->seq_info[park_pos]->id_num = qs->q_num;
		if(len < qs->min_len){
			qs->min_len = len;
		}
		
		if(len > qs->max_len){
			qs->max_len = len;
		}		
		reverse_complement(qs->seq_info[park_pos]->seq ,rc,len);
		qs->seq_info[park_pos]->reverse_seq = qs->string + string_pos;
		for(i = 0; i < len;i++){
			qs->string[string_pos] = rc[i];
			string_pos++;
		}
		qs->string[string_pos] = 0;
		//string_pos++;
	}
	qs->size = size;
	return FILE_END;

}

struct qs_struct* search_fasta_fastq(struct qs_struct* qs,struct parameters* param) 
{
	FILE *file;
	unsigned int string_pos = 0;
	int park_pos = -1;
	char line[MAX_LINE];
	int i,j,c,n;
	int seq_p = 0;
	int set = 0;
	long int len = 0;
	long int total_len = 0;
	unsigned char* chromsome =  malloc(sizeof(unsigned char)* 300000000);
	
	long int* lengths = 0; 
	lengths = malloc(sizeof(unsigned int) * 50);
	unsigned char** chr_names = 0;
	chr_names = malloc(sizeof(unsigned char*) * 50);
	
	
	if (!(file = fopen(param->infile[2], "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->infile[2]);
		exit(-1);
	}
	c= -1;
	
	while(fgets(line, MAX_LINE, file)){
		//reading name .... 
		if((line[0] == '@' && !set)|| (line[0] == '>' && !set)){
			//set sequence length of previous read
			
			if(park_pos != -1){ // to avoid setting something before the first read was read
				
				qs =  search_hash(qs,chromsome,len,param->r , total_len);
				for(i = 0; i< qs->size;i++){
					qs->seq_info[i]->last_f_test = -1000;
					qs->seq_info[i]->last_r_test = -1000;
				}
				total_len += len;
				//fprintf(stderr,"	%d	%ud	%ud	(%f)\n",len,mers,total_len,(float)mers/(float)total_len*100.0f);
			}
			//fprintf(stderr,"%s	%ld",line,total_len);
			
			n = strlen(line);
			c++;
			
			chr_names[c] = malloc(sizeof(unsigned char) * (n+1));
			for(i = 1; i < n-1;i++){
				chr_names[c][i-1] = line[i];
			}
			chr_names[c][n-1] = 0; 
			lengths[c] = len;
			
			string_pos = 0;			
			len = 0;
			seq_p = 1;
			set = 1;			
			park_pos++;
			//get ready to read quality if present  
		}else if(line[0] == '+' && !set){
			seq_p = 0;
			set = 1;
			//reading sequence or quality  
		}else{			
			if(!set){
				string_pos--;
			}
			
			if(seq_p){
				for(i = 0;i < MAX_LINE;i++){
					if(iscntrl((int)line[i])){
						chromsome[string_pos] = 0;
						//qt->string[string_pos] = 0;
						string_pos++;
						break;
					}
					
					len++;
					chromsome[string_pos] = nuc_code5[(int)line[i]];
					string_pos++;
				}
			}
			set = 0;
			
		}
	}
	c++;

	lengths[c] = len;
	chr_names[c] = 0;
	
	for(i = 1; i <= c;i++){
		lengths[i] +=  lengths[i-1];
	}
	qs =  search_hash(qs,chromsome,len,param->r , total_len);

	//for(i = 0; i <= c;i++){
	///	fprintf(stderr,"%s	%ld\n",chr_names[i],lengths[i]);
	//}
	//qs = search_chromosome_hash(qs,param,chromsome,len,total_len);
	//total_len +=  len;
	
	//fwrite(&param->upper_limit, sizeof(unsigned int),1,binfile);
	for(i = 0; i< qs->size;i++){
		print_sam(qs->seq_info[i], chr_names, lengths ,c);
	}
	
	for(i = 0; i< qs->size;i++){
		for(j = 0; j < LIST_STORE_SIZE;j++){
			qs->seq_info[i]->match_units[j] = 0ul;
		}
	}
	//string_pos = 0;
	qs->q_num = 0;
	
	for(i = 0; i < c;i++){
		free(chr_names[i]);
	}
	
	free(chr_names);
	free(lengths);
	free(chromsome);
	fclose(file);
	return qs;
}

void print_sam(struct seq_info* si, unsigned char** t_names, long int* lengths,int size)
{
	char alphabet[] = "ACGTN";
	char* pline = malloc(sizeof (char)* MAX_LINE);
	char seq[MAX_LINE];
	unsigned char star[] = "*";
	int strand = 0;
	int no_qual = 0;
	int i,j;
	unsigned char* t_name =  0;
	long int pos = 0;
	int matches = 0;
	
	pline[0] = 0 ;
	strand = si->match_units[0] & 1ul;
	if(strand){
		strand = 16;
	}
	
	//fprintf(stderr,"before : %s\n",si->qual);
	
	if(!si->qual){
		si->qual = star;
		no_qual = 1;
	}
	
	//fprintf(stderr,"after : %s\n",si->qual);
	
	if(!si->match_units[0]){
		
		
		
		for(i = 0;i < si->len;i++){
			seq[i] = alphabet[si->seq[i]];
		}
		seq[si->len] = 0;
		
		sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
			si->sn ,//1 QNAME Query NAME of the read or the read pair 
			4,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
			"*",//3 RNAME Reference sequence NAME 
			0,//4 POS 1-based leftmost POSition of clipped alignment 
			0,//5 MAPQ MAPping Quality (Phred-scaled) 
			"*",//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
			"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
			0,//8 MPOS 1-based leftmost Mate POSition 
			0,//9 ISIZE inferred Insert SIZE 
			seq,//10 SEQ query SEQuence on the same strand as the reference 
			si->qual//11 QUAL query QUALity (ASCII-33=Phred base quality) 
			);
		i = strlen(pline);
		pline[i] = 0;
		fprintf(outfile, "%s\n",pline);
	}else{
		pos =(long int) ((si->match_units[0] >> 1ul )& 0xFFFFFFFFul);
		matches = (int) (si->match_units[0] >> 56ul);
		//fprintf(stderr,"printing:%lu	%ld\n",si->match_units[0],pos);
		j = 0;
		for(i = 0; i <= size;i++){
			if( pos < lengths[i]){
				t_name = t_names[i-1];
				pos = pos - lengths[i-1];
				j = i-1 + 1;
				break;
			}
		}
		if(!j){
			fprintf(stderr,"not found: %ld\n",pos);
			for(i = 0; i <= size;i++){
				fprintf(stderr,"%s	%ld	%s\n", t_names[i],lengths[i],si->sn);
			}
			exit(-1);
		}
		//fprintf(stderr,"NAME found : %d\n",j-1);
		
		if(!strand){
			for(i = 0;i < si->len;i++){
				seq[i] = alphabet[si->seq[i]];
			}
		}else{
			for(i = 0;i < si->len;i++){
				seq[i] = alphabet[si->reverse_seq[i]];
			}
		}
		
		seq[si->len] = 0;
		//fprintf(stdout, "LINe:\n\n");
		if(!strand){
			sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
				si->sn ,//1 QNAME Query NAME of the read or the read pair 
				strand,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
				t_name,//3 RNAME Reference sequence NAME 
				(int)(pos+1l),//4 POS 1-based leftmost POSition of clipped alignment 
				255,//5 MAPQ MAPping Quality (Phred-scaled) 
				"*",//si->len,//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
				"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
				0,//8 MPOS 1-based leftmost Mate POSition 
				0,//9 ISIZE inferred Insert SIZE 
				seq,//10 SEQ query SEQuence on the same strand as the reference 
				si->qual//11 QUAL query QUALity (ASCII-33=Phred base quality) 
				);
		}else{
			if(no_qual){
				sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
						si->sn ,//1 QNAME Query NAME of the read or the read pair 
						strand,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
						t_name,//3 RNAME Reference sequence NAME 
						(int)(pos),//4 POS 1-based leftmost POSition of clipped alignment 
						255,//5 MAPQ MAPping Quality (Phred-scaled) 
						"*",//si->len,//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
						"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
						0,//8 MPOS 1-based leftmost Mate POSition 
						0,//9 ISIZE inferred Insert SIZE 
						seq,//10 SEQ query SEQuence on the same strand as the reference 
						si->qual//11 QUAL query QUALity (ASCII-33=Phred base quality) 
						);
			}else{
				sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
						si->sn ,//1 QNAME Query NAME of the read or the read pair 
						strand,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
						t_name,//3 RNAME Reference sequence NAME 
						(int)(pos),//4 POS 1-based leftmost POSition of clipped alignment 
						255,//5 MAPQ MAPping Quality (Phred-scaled) 
						"*",//si->len,//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
						"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
						0,//8 MPOS 1-based leftmost Mate POSition 
						0,//9 ISIZE inferred Insert SIZE 
						seq,//10 SEQ query SEQuence on the same strand as the reference 
						reverse_without_complement(si->qual, si->len)//11 QUAL query QUALity (ASCII-33=Phred base quality) 
						);
			}
		}
		
		i = strlen(pline);
		pline[i] = '\t';
		i++;
		sprintf(pline+i,"NM:i:%d",  si->len - matches);
		
		
		if(si->match_units[1]){
			
			i = strlen(pline);
			pline[i] = '\t';
			i++;
			sprintf(pline+i,"XA:Z:");
			j = 1;
			while( j < LIST_STORE_SIZE){
				if(! si->match_units[j] ){
					break;
				}
				strand = si->match_units[j] & 1ul;
				pos = (unsigned int)  ((si->match_units[j] >> 1ul )& 0xFFFFFFFFul);
				for(i = 0; i <= size;i++){
					//fprintf(stderr," looking at: %d	%d	%d	%s\n",i,pos,lengths[i],t_names[i]);
					if( pos < lengths[i]){
						t_name = t_names[i-1];
						pos = pos - lengths[i-1];
						break;
					}
				}
				
				i = strlen(pline);
				if(!strand){
					sprintf(pline+i,"%s,+%d,%dM,%d;", t_name,(int)(pos+1l),si->len, si->len - (int)(si->match_units[j] >> 56ul));
				}else{
					sprintf(pline+i,"%s,-%d,%dM,%d;", t_name,(int)(pos+1l),si->len, si->len - (int)(si->match_units[j] >> 56ul));
				}
				j++;
			}
			
		}
		
		i = strlen(pline);
		pline[i] = 0;
		fprintf(outfile, "%s\n",pline);
	}
	if(no_qual){
		si->qual = 0;
	}
	
	free(pline);
}


void reverse_complement(unsigned char* p,unsigned char* rc,int len)
{
	int rev[5] = {3,2,1,0,4};
	
	int i,c;
	c = 0;
	for(i =len-1; i >= 0;i--){
		rc[c] = rev[(int)p[i]];
		c++;
	}
	rc[c] = 0;
}

struct qs_struct* make_hash(struct qs_struct* qs,int r)
{
	unsigned int key = 0;
	int i,j,c;
	int elements = 0;
	unsigned char* tmp = 0;
	int m;
	int n = qs->size;
	
	qs->hash_index = malloc(sizeof(int) * ((1 << r * 2) +1));
	
	for(i = 0; i < (1 <<r * 2)+1;i++){
		qs->hash_index[i] = 0;
	}
	
	for(i = 0; i < n;i++){
		m = qs->seq_info[i]->len;
		if(m > 31){
			m = 31;
		}
		tmp = qs->seq_info[i]->seq;//   sequences[i];
		for(j = 0;j + r  <= m;j+= r){
			key = 0;
			for(c = 0;c < r;c++){
				key = (key << 2u) | ((int)tmp[c]   & 0x3u);			
			}
			qs->hash_index[key+1]++;
			tmp += r;
		}	
		//m = qs->seq_info[i]->len;
		tmp = qs->seq_info[i]->reverse_seq;//   sequences[i];
		for(j = 0;j + r  <= m;j+= r){
			key = 0;
			for(c = 0;c < r;c++){
				key = (key << 2u) | ((int)tmp[c]& 0x3u);		
			}
			qs->hash_index[key+1]++;
			tmp += r;
		}	
	}
	
	for(i = 1; i < (1 << r * 2)+1;i++){
		qs->hash_index[i] +=  qs->hash_index[i-1];
	}
	elements = qs->hash_index[(1 << r * 2) ];
	
	qs->hash = malloc(sizeof(int) * (elements));
	for(i = 0; i < elements;i++){
		qs->hash[i] = -1;
	}
	for(i = 0; i < n;i++){
		m = qs->seq_info[i]->len;
		if(m > 31){
			m = 31;
		}
		tmp = qs->seq_info[i]->seq;//   sequences[i];
		for(j = 0;j + r  <= m;j+= r){
			key = 0;
			for(c = 0;c < r;c++){
				key = (key << 2u) | ((int)tmp[c]   & 0x3u);			
			}
			for(c = qs->hash_index[key];c < qs->hash_index[key+1u];c++){
				if(qs->hash[c] == -1){
					qs->hash[c] = (i << 8u) | j;
					//qs->hash[c] = i;
					break;
				}
			}
			tmp += r;
		}	
		
		
		
		//m = qs->seq_info[i]->len;
		tmp = qs->seq_info[i]->reverse_seq;//   sequences[i];
		
		
		
		for(j = 0;j + r  <= m;j+= r){
			key = 0;
			for(c = 0;c < r;c++){
				key = (key << 2u) | ((int)tmp[c]& 0x3u);
			
			}
			for(c = qs->hash_index[key];c < qs->hash_index[key+1u];c++){
				if(qs->hash[c] == -1){
					qs->hash[c] = (i << 8u) | j | 0x80u;
					break;
				}
			}
			tmp += r;
		}	
	}
	//exit(0);
	return qs;
}

struct qs_struct* search_hash(struct qs_struct* qs,unsigned char* t,int n, long int r,long int offset)
{
	int j,g;
	int i;
	unsigned char* target = 0;
	struct seq_info* si = 0;
	unsigned int key = 0;
	unsigned int mask = 0;
	int t_len = 0;
	long int seed_offset = 0;
	int strand = 0;
	unsigned long int new = 0;
	//static int test = 0;
	long int add = 0;

	for(i = 0; i <  r;i++){
		mask = (mask << 2u) | 3u;
	}
	//timer(0,0);
	key = 0u;
	for(i = 0; i <  r-1;i++){
		key = (key << 2u) | (t[i] & 0x3u);
	}
	for(i = r-1; i < n;i++){
		key = ((key << 2u) | (t[i] & 0x3u)) & mask;
		if(key){
			g = 0;
			for(j = qs->hash_index[key];j < qs->hash_index[key+1u];j++){
				seed_offset = qs->hash[j] & 0x7fu;
				
				si = qs->seq_info[ qs->hash[j] >> 8u];
				//id = qs->hash[j] >> 8u;
				strand = qs->hash[j] & 0x80u;
				add = ( (long int)i + ((long int) si->len - seed_offset)  -r + 5l);
				if( add > 0){
					target = t +add;  
				}else{
					target = t;
					add = 0;
				}
				
				
				t_len = si->len + 10;
				if( add <= n){
				if(!strand){
					if(si->last_f_test <= i){
						new = validate_bpm(target ,si->seq , t_len , si->len, offset+ add);
						BinaryInsertionSortOne(si->match_units,LIST_STORE_SIZE ,new);
						si->last_f_test = i + si->len  ;
					}
				}else{
					if(si->last_r_test <= i){
						new = validate_bpm(target ,si->reverse_seq , t_len , si->len,offset + add);
						new |= 1ul ;
						BinaryInsertionSortOne(si->match_units,LIST_STORE_SIZE,new);
						si->last_r_test = i + si->len ;
					}
				}
				}
			}
		}
		//if(i % 100000 == 0){
		//	fprintf(stderr,"\r%8.2f percent done",(double)i /(double)n * 100);
		//}
	}

	//fprintf(stderr,"\r%8.2f percent done\n",100.0);
	//exit(0);
	
	//test++;
	//timer(qs->size,n);
	return qs;
}



unsigned int rev_bits(unsigned int x)
{
	x = ((x & 0x55555555) << 1) | ((x >>1) & 0x55555555);
	x = ((x & 0x33333333) << 2) | ((x >>2) & 0x33333333);
	x = ((x & 0x0F0F0F0F) << 4) | ((x >>4) & 0x0F0F0F0F);
	x = (x << 24) | ((x & 0xFF00) << 8) | ((x >>8) & 0xFF00) | (x >> 24);
	return x;	
}










