/*
 
 Copyright (C) 2010 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#include "delve.h"
#include "seed.h"
#include "interface.h"
#include "rand.h"
#include "dbio.h"
#include "simulate.h"
#include "read_sam.h"
#include  <math.h>
#include <string.h>


void simulate(struct parameters* param)
{
	//unsigned long int fpos = 0u;
	struct qs_struct* qs = 0;
	int i;
	FILE *fastafile = 0; 
	qs = init_query_structure(qs,param);
	
	float prob[5] = {0.01f,0.03f,0.9f,0.0f,0.0f};
	for(i = 2; i < param->infiles;i++){
		prob[i-2] = atof(param->infile[i]);
	}
	
	for(i = 0;i < 5;i++){
		if(prob[i] < 0.0f || prob[i] > 1.0f){
			fprintf(stderr,"Probabilities have to be 0 <= p <= 1\n");
			exit(1);
		}
	}
	if (!(fastafile = fopen(param->infile[1], "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->infile[1]);
		exit(-1);
	}
	qs->size = param->num_query;
	
	//while(fpos != FILE_END){
	//	fpos = read_fasta_fastq(fastafile,qs,param,fpos);
	while(qs->size == param->num_query){
		qs = read_fasta_fastq2(fastafile,qs,param);
		qs = mutate_reads(qs,prob);
	
	}
	
	fclose(fastafile);
	free_query_structure(qs,param);
	
}

struct qs_struct* mutate_reads(struct qs_struct* qs,float* prob)
{
	unsigned char* tmp_seq = 0;
	unsigned char* tmp_qual = 0;
	int n = qs->size;
	int i,j,c;
	int m = 0;
	float a,b;
	char alphabet[] = "ACGTN";
	
	
	tmp_seq = malloc(sizeof(unsigned char) * MAX_LINE);
	tmp_qual = malloc(sizeof(unsigned char) * MAX_LINE);
	for(i = 0; i < n;i++){
		m = qs->seq_info[i]->len;
		b  = ((prob[1] - ((float)m * prob[0])) / (float)(m -1)) * -1.0f;
		
		a = prob[0] - b;
		
		//fprintf(stderr,"\n%f	%f\n y = %f * x + %f\n",prob[0],prob[1],a,b);
		c = 0;

		if(ran1(seed) <= prob[3]){
			tmp_seq[c] = 'G';
			tmp_qual[c] = qs->seq_info[i]->qual[0]; 
			c++;
		}
		
		for(j = 0; j < m;j++){
		//	fprintf(stderr,"%d	%f\n",j, (float)(j+1)* a +b);
			if(ran1(seed)  <= (float)(j+1)* a +b){
				if(ran1(seed) <= prob[2]){
					tmp_seq[c] = alphabet[random_int(4)];
					tmp_qual[c] = qs->seq_info[i]->qual[j];
					c++;
					//fprintf(stderr,"MISMATCH\n");
				}else{
					if(ran1(seed) <= 0.5f){
						tmp_seq[c] = alphabet[random_int(4)];
						tmp_qual[c] = qs->seq_info[i]->qual[j];
						c++;
						tmp_seq[c] = alphabet[qs->seq_info[i]->seq[j]];
						tmp_qual[c] = qs->seq_info[i]->qual[j];
						c++;
					//	fprintf(stderr,"INSERT\n");
					}else{
					//	fprintf(stderr,"DELETION\n");
					}
				}
			}else{
				tmp_seq[c] = alphabet[qs->seq_info[i]->seq[j]];
				tmp_qual[c] = qs->seq_info[i]->qual[j];
				c++;
			}
		}
		if(ran1(seed) <= prob[3]){
			tmp_seq[c] = 'A';
			tmp_qual[c] = qs->seq_info[i]->qual[m-1]; 
			c++;
		}
		
		tmp_seq[c]  = 0;
		tmp_qual[c] = 0;
		
		
		/*for(j = 0; j < m;j++){
			fprintf(stderr,"%c",alphabet[qs->seq_info[i]->seq[j]]);
		}
		fprintf(stderr,"\n%s\n\n",qs->seq_info[i]->qual);*/
		
		//fprintf(stderr,"%s\n%s\n%s\n", qs->seq_info[i]->sn , tmp_seq,tmp_qual);
		fprintf(outfile,"@%s\n%s\n+\n%s\n",qs->seq_info[i]->sn , tmp_seq,tmp_qual);
		
	}
	free(tmp_seq);
	free(tmp_qual);
	return qs;
}


void simulate_reads_from_bed(struct db* db,struct parameters* param)
{
	FILE* file = 0;
	FILE* file2 = 0;
	char line[MAX_LINE];
	char seq_name[MAX_LINE]; 
	unsigned char seq[MAX_LINE]; 
	int num_sequences = param->sim_numseq;
	int column = 0; 
	int i,j,tmp;
	unsigned int start;
	unsigned int stop;
	unsigned int chr_start;
	int strand = 0;
	int c = 0;
	int hit = 0;
	int rank = 0;
	float y = 0.0f;
	float total = 0.0f;
	int errors;
	
	int* rank_arr = 0;
	
	unsigned int sim_start; 
	
	
	if(param->control_file){
		if (!(file2 = fopen(param->control_file , "w" ))){
			fprintf(stderr,"Cannot open bed file '%s'\n",param->control_file);
			exit(-1);
		}
	}
	
	if (!(file = fopen(param->infile[1] , "r" ))){
		fprintf(stderr,"Cannot open bed file '%s'\n",param->infile[1]);
		exit(-1);
	}
	
	while(fgets(line, MAX_LINE, file)){
		if(line[0] != '#'){
			total++;
		}
	}
	rewind(file);
	
	rank_arr = malloc(sizeof(int) * total);
	
	
	for(i = 0; i < total;i++){
		rank_arr[i] = i+1;
	}
	rank_arr = shuffle_arr(rank_arr ,total);
	
	//fprintf(stderr,"TOTAL:%f\n",total);
	//exit(0);
	while (c < num_sequences){
		
		while(fgets(line, MAX_LINE, file)){
			if(line[0] != '#'){
				column = 1; //<QNAME> 
				tmp = 0;
				hit = 0;
				for(j = 0;j < MAX_LINE;j++){
					tmp++;
					if(isspace((int)line[j])){
						break;
					}
				}
				
				for(j = 0;j < MAX_LINE;j++){
					
					if(isspace((int)line[j])){
						seq_name[j] = 0;
						break;
					}
					seq_name[j] = line[j];
				}
				chr_start = get_chr_offset(db->seq_info, seq_name);
				//fprintf(stderr,"chr:%s	%u\n",seq_name,chr_start);
				start = 0;
				stop = 0;
				for(i = 0; i < MAX_LINE;i++){
					if(line[i] == '\n'){
						break;
					}
					if(isspace((int)line[i])){
						column++;
						switch(column){
							case 2: // <FLAG>
								start  = chr_start + atoi(line+i+1);
								break;
							case 3: // <RNAME> 
								stop = chr_start + atoi(line+i+1);
								break;
							case 6: // <RNAME> 
								if(*(line+i+1) == '+'){
									strand = 0;
								}else{
									strand = 1;
								}
								break;
						}
					}
				}
				//fprintf(stderr,"%s",line);
				
				
				//y = x / (x^(1.2+  (x - floor(x/1000) *1000 )/ 1000 ))
				
				y = (float)rank_arr[rank] / (  pow((float) rank_arr[rank], 1.2f + ((float) (rank_arr[rank] % (int)(total)) / total)));
			
				y = y * 100 + random_int( (int)((y * 100) / 10) + 2 );
				//y = 10;
				//fprintf(stderr,"%s	%u	%u	%d	=>reads: %d	%d\n",seq_name,start,stop,strand,(int)(y), random_int(2) );
				for(j = 0; j < (int)(y);j++){
					
					
					if(!param->window_start){
						sim_start = (unsigned int) (random_int((int)((stop-(param->sim_length+1))  - start)));
						sim_start += start;
					//	fprintf(stderr,"%u	%u	%u	%u	%u\n",start, stop,sim_start,sim_start + start ,param->sim_length);
						i = get_sequence(db,seq,sim_start,param->sim_length,strand);
					}else{
						sim_start = (unsigned int) (random_int((int)(2*param->window_start)));
						
						if(!strand){
							sim_start = sim_start - param->window_start;
							sim_start += start;
							
							i = get_sequence(db,seq,sim_start,param->sim_length,strand);
						}else{
							sim_start =  sim_start - param->window_start;
							sim_start += stop;
							i = get_sequence(db,seq,sim_start  ,param->sim_length,strand);
						}
					}
					
					//
					//fprintf(stderr,"%u	%u	%u\n",sim_start,sim_start + start ,param->sim_length);
					
					
					//for(i = 0; i < param->sim_length;i++){
					//	fprintf(stderr,"%c",alphabet[seq[i]]);
					//}
					//fprintf(stderr,"\n");
					
					if(param->control_file){
						print_sequence(file2,seq,sim_start,strand,param);
					}
					
					errors = mutate_sequence(seq,param);
					//fprintf(stderr,"%s	%d\n", seq,c);	
					
					if(!strand){
						fprintf(stdout,">F%X_%d\n%s\n",sim_start,errors ,seq);
					}else{
						fprintf(stdout,">B%X_%d\n%s\n",sim_start,errors,seq);
					}
					//c = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
					c++;
					if(c == num_sequences){
						break;
					}
				}
				rank = rank + 1;
				if(c == num_sequences){
					break;
				}
			}
		}
		rank = 0;
		rewind(file);
	}
	fclose(file);
	if(param->control_file){
		fclose(file2);
	}
	free(rank_arr);
}





struct db* evaluate_sam_file(struct db* db,struct parameters* param)
{
	FILE* file = 0;
	char tmp_text[MAX_LINE];
	char line[MAX_LINE];
	char command[1000];
	//char seq_name[MAX_LINE]; 
	int column = 0; 
	int i,j;
	int tmp = 0;
	unsigned int pos;
	unsigned int add = 0;
	//int read = 0;
	//int c = 0;
	//int hit = 0;
	
	
	int storage_num_chr = 10; 
	db = malloc(sizeof(struct db));
	
	db->seq_info = malloc(sizeof(struct db_seq_info));
	db->seq_info->names = malloc(sizeof(char*)* storage_num_chr);
	db->seq_info->sn_len = malloc(sizeof(int)* storage_num_chr);
	db->seq_info->len = malloc(sizeof(unsigned int)* storage_num_chr);
	db->seq_info->nuc_counts = malloc(sizeof(float)*5);
	
	db->bfa = 0;
	db->dbname  = malloc(sizeof(char)*(2));
	db->dbname[0] = 'A';
	db->dbname[1] = 0;
	
	int read_seq_names = 0;
	
	int real_strand = 0;
	int read_strand = 0;
	unsigned int real_start = 0;
	int Q = 0;
	int k_distance = 0;
	
	int sam = 0;
	int total[11];
	int errors[11];
	
	int real_errors = 0;
	
	int k_distances_in_error[11][5];
	
	int error_types[6][6];
	
	for(i = 0; i < 6;i++){
		for(j = 0; j < 6;j++){
			error_types[i][j] = 0;
		}
	}
	
	int KR[5]; 
	int KA[5];
	for(i =0;i < 5;i++){
		KR[i] = 0;
		KA[i] = 0;
	}
	
	int exclude = 0;
	for(i = 0; i < 11;i++){
		total[i] = 0;
		errors[i] = 0;
		for(j = 0; j < 5;j++){
			k_distances_in_error[i][j] = 0;
		}
	}
	
	
	
	if(!strcmp(".sam", param->infile[1] + (strlen(param->infile[1] ) - 4))){
		sam = 1;
		//}else if (byg_end(".bam", param->infile[i])  == strlen(param->infile[i])){
	}else if (!strcmp(".bam", param->infile[1] + (strlen(param->infile[1] ) - 4))){
		sam = 2;
		//}else if (byg_end(".fa", param->infile[i])  == strlen(param->infile[i])){
	}
		
	
	
	if(!sam){
		fprintf(stderr,"eval only works with sam / bam files.\n");
	}
	if(sam == 2){
		command[0] = 0;
		strcat ( command, "samtools view -H ");     // Copy name into full name
		i = sprintf (tmp_text, "%s ", param->infile[1]);
		strcat ( command, tmp_text);
		//fprintf(stderr,"%s\n",command);
		if (!(file = popen(command, "r"))) {
			fprintf(stderr,"Cannot open bam file '%s' with command:%s\n",param->infile[1],command);
			exit(-1);
		}
	}else{
		if (!(file = fopen(param->infile[1] , "r" ))){
			fprintf(stderr,"Cannot open sam file '%s'\n",param->infile[1]);
			exit(-1);
		}
		
	}
	
	
	
	
	while(fgets(line, MAX_LINE, file)){
		if(line[0] == '@'){
			if(line[1] == 'S' && line[2] == 'Q' ){
			
				tmp = byg_end("SN:", line );
				if(tmp){
					db->seq_info->names[read_seq_names] = malloc(sizeof(unsigned char)*(50) );
					add = 0;
					for(j = tmp;j < MAX_LINE;j++){
						if(isspace((int)line[j])){
							db->seq_info->names[read_seq_names][add] = 0;
							break;
						}
						db->seq_info->names[read_seq_names][add] = line[j];
						add++;
					}
				}
				
				//db->seq_info->names[read] = (unsigned char*)strdup(name);
				db->seq_info->sn_len[read_seq_names] = add;
				tmp = byg_end("LN:", line );
				db->seq_info->len[read_seq_names] = atoi(line+tmp);
				read_seq_names++;
				if(read_seq_names == storage_num_chr){
					storage_num_chr = storage_num_chr << 1;
					db->seq_info->names = realloc(db->seq_info->names,sizeof(char*)* storage_num_chr);
					db->seq_info->sn_len = realloc(db->seq_info->sn_len,sizeof(int)* storage_num_chr);
					db->seq_info->len = realloc(db->seq_info->len,sizeof(int)* storage_num_chr);
				}
			}
			
		}else{
			break;
		}
	}
	if(!read_seq_names){
		fprintf(stderr,"SAM header seems to be absent.\n");
		fclose(file);
		return db;
		
	}
	
	db->seq_info->num_seq = read_seq_names;
	//real_start = 0;
	pos = 0;
	for(i = 0; i < db->seq_info->num_seq ;i++){
		real_start = db->seq_info->len[i];
		db->seq_info->len[i] = pos;
		
		pos = real_start + pos;
		
		//fprintf(stderr,"%s	%u\n",db->seq_info->names[i],db->seq_info->len[i]);
	}
	
	//real_start = 0;
	
	if(sam == 2){
		pclose(file);
		command[0] = 0;
		strcat ( command, "samtools view  ");     // Copy name into full name
		i = sprintf (tmp_text, "%s ", param->infile[1]);
		strcat ( command, tmp_text);
		//fprintf(stderr,"%s\n",command);
		if (!(file = popen(command, "r"))) {
			fprintf(stderr,"Cannot open bam file '%s' with command:%s\n",param->infile[1],command);
			exit(-1);
		}
	}else{
		rewind(file);		
	}
	
	
	
	
	while(fgets(line, MAX_LINE, file)){
		if(line[0] != '@'){
			column = 1; //<QNAME> 
			pos = 0;
			real_errors = -1;
			exclude = 0;
			if(line[0] == 'F'){
				real_strand = 0;
			}else{
				real_strand = 0x10;
			}
			real_start = 0;
			for(j = 1;j < MAX_LINE;j++){
				if(line[j] == '_'){
					real_errors = atoi(line+j+1);
					if(real_errors > 4){
						real_errors = 4;
					}
					
				}
				if(isspace((int)line[j])){
				//	ri[c]->name[j] = 0;
					break;
				}
				if(real_errors == -1){
					add = 0;
					switch (line[j]) {
						case '0':
							add = 0;
							break;
						case '1':
							add = 1;
							break;
						case '2':
							add = 2;
							break;
						case '3':
							add = 3;
							break;
						case '4':
							add = 4;
							break;
						case '5':
							add = 5;
							break;
						case '6':
							add = 6;
							break;
						case '7':
							add = 7;
							break;
						case '8':
							add = 8;
							break;
						case '9':
							add = 9;
							break;
						case 'A':
							add = 10;
							break;
						case 'B':
							add = 11;
							break;
						case 'C':
							add = 12;
							break;
						case 'D':
							add = 13;
							break;
						case 'E':
							add = 14;
							break;
						case 'F':
							add = 15;
							break;
						default:
							break;
					}
					
					real_start = (real_start << 4u) +  add;
				}
				//ri[c]->name[j] = line[j];
			}
			
			for(i = 0; i < MAX_LINE;i++){
				if(line[i] == '\n'){
					break;
				}
				if(isspace((int)line[i])){
					column++;
					switch(column){
						case 2: // <FLAG>
							tmp = atoi(line+i+1);
							read_strand = (tmp & 0x10);
							if(tmp == 4){
								exclude = 1;
							}
						//	ri[c]->strand[hit] = (tmp & 0x10);
							break;
						case 3: // <RNAME> 
							if(line[i+1] != '*'){
								tmp = 0;
								for(j = i+1;j < MAX_LINE;j++){
									if(isspace((int)line[j])){
										tmp_text[tmp] = 0;
										break;
									}
									tmp_text[tmp] = line[j];
									tmp++;
								}
								pos = get_chr_offset(db->seq_info, tmp_text);
								
								//fprintf(stderr,"%s	%u\n",tmp_text,pos);
								if(pos == 0xFFFFFFFFu){
									fprintf(stderr,"ERROR: %s not found in indexed sequences\n",tmp_text);
									fprintf(stderr,"%s\n",line);
									exit(-1);
								}
							}
							break;
						case 4: // <POS>
							if(pos != 0xFFFFFFFFu){
								//fprintf(stderr,"Adding %u	%u	= %u\n",pos,(unsigned int) atoi(line + i + 1),(unsigned int) atoi(line + i + 1) + pos);
								pos += (unsigned int) atoi(line + i + 1)  - 1;
								
							}
							break;
						case 5: //  <MAPQ>
							Q = atoi(line +i +1) / 10;
							if(Q > 10){
								Q = 10;
							}
							//ri[c]->mapq =  1.0 - pow(10.0, -1.0 *   atof(line +i +1) / 10.0); 
							
							break;
						
					}
				}
			}
			
			
			k_distance = 0;
			i = byg_end("NM:i:", line );
			if(i){
				k_distance = atoi(line+i);
				if(k_distance > 4){
					k_distance = 4;
				}
				KA[k_distance]++;
			}
			
			KR[real_errors]++;
			
			
			if( k_distance > param->error_cutoff){
				exclude = 1;
			}
			
			if(exclude){
				error_types[k_distance][5]++;
			}
			
			if(!exclude){
				total[Q]++;				
				if(real_strand == read_strand){
					if(abs((long int)real_start - (long int) pos) < 10){ 
						if(k_distance == real_errors){
							error_types[k_distance][0]++;
							//fprintf(stdout,"CORRECT\n");
						}else{
							error_types[k_distance][1]++;
							//fprintf(stdout,"WRONG NUC\n");
							fprintf(stdout,"%s",line);
						}
						
						
						//fprintf(stderr,"%d	%d\n",real_strand,read_strand);
						//fprintf(stderr,"%u	%u	%d\n",real_start,pos,Q);
					}else{
						//fprintf(stdout,"%s",line);
						if(k_distance == real_errors){
							error_types[k_distance][2]++;
							
						}else if(k_distance < real_errors ){
							error_types[k_distance][3]++;
						}else{
							error_types[k_distance][4]++;
						}
						
						errors[Q]++;
						k_distances_in_error[Q][k_distance]++;
					}
				}else{
					//fprintf(stdout,"%s",line);
					
					if(k_distance == real_errors){
						error_types[k_distance][2]++;
					}else if(k_distance < real_errors ){
						error_types[k_distance][3]++;
					}else{
						error_types[k_distance][4]++;
					}
					
					
					errors[Q]++;
					k_distances_in_error[Q][k_distance]++;
				}
			}
		}
	}
	
	
	if(sam==2){
		pclose(file);
	}else{
		fclose(file);
	}
	//fclose(file);
	
	fprintf(stderr,"K\tT\tTPW \tKa=Ks\tKa<Ks\tKa>Ks\tNA\tCT\tCEP\tT\tTPW \tKa=Ks\tKa<Ks\tKa>Ks\tNA\tKs\tKa\tKs\tKa\n");
	for(i = 0; i < 5;i++){
		if(i < 5){
			fprintf(stderr,"%d:\t",i);
		}else{
			fprintf(stderr,"Total:\t");
		}
		add = 0;
		for(j = 0; j < 6;j++){
			
			//if(i != 5){
			//	error_types[5][j] += error_types[i][j];
			//}
			if(i){
				error_types[i][j] = error_types[i][j] + error_types[i-1][j];
			}
			fprintf(stderr,"%d\t",error_types[i][j]);
			add += error_types[i][j];
			
		}
		
		fprintf(stderr,"%d\t%0.2f%%\t",add - error_types[i][5],100.0f - (float)(error_types[i][0] + error_types[i][1]) / (float)(add - error_types[i][5])*100.0f);
		
		for(j = 0; j < 6;j++){
			fprintf(stderr,"%0.2f%%\t",(float)error_types[i][j]/(float)add  * 100.0f);
		}
		fprintf(stderr,"%d\t%d\t%0.2f%%\t%0.2f%%",KR[i],KA[i],(float) KR[i] / (float)(KR[0]+KR[1]+KR[2]+KR[3]+KR[4])*100.0f,(float) KA[i] / (float)(KA[0]+KA[1]+KA[2]+KA[3]+KA[4])*100.0f);
		fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
	
	fprintf(stderr,"Q\tErrors\tTotal\tError %%\tCE\tCT\tCEP\n");
	for(i = 10;i >= 0;i--){
		if(total[i]){
			tmp = i;
			break;
		}
	}
	
	for(i = tmp;i >= 0;i--){
		if(total[i]){
			fprintf(stderr,"%d\t%d\t%d\t%0.2f%%\t",i*10,errors[i],total[i],(float) errors[i]/ (float)total[i]*100.0f);
		}else{
			fprintf(stderr,"%d\t%d\t%d\t%0.2f%%\t",i*10,errors[i],total[i],(float) 0.0);
		}
		if(i != 10){
			total[i] += total[i+1];
			errors[i] += errors[i+1]; 
			
			for(j = 0;j <5;j++){
				k_distances_in_error[i][j] += k_distances_in_error[i+1][j]; 
			}
			
		}
		fprintf(stderr,"%d\t%d\t%0.2f%%\t",errors[i],total[i],(float) errors[i]/ (float)total[i]*100.0f);
		for(j =0; j < 4;j++){
			fprintf(stderr,"%d\t",k_distances_in_error[i][j] );
		}
		fprintf(stderr,"%d\n",k_distances_in_error[i][4] );
	}
	return db;
}



void print_sequence(FILE* file,unsigned char* seq,unsigned int sim_start,int strand,struct parameters*  param)
{
	
	int j;
	char tmp_seq[MAX_LINE]; 
	char alphabet[] = "ACGTN";
	
	for(j = 0; j < param->sim_length;j++){
		tmp_seq[j] = alphabet[seq[j]];
		//j++;
	}
	tmp_seq[j] = 0;
	if(!strand){
		fprintf(file,">F%X\n%s\n",sim_start ,tmp_seq);
	}else{
		fprintf(file,">B%X\n%s\n",sim_start,tmp_seq);
	}
	
}



int mutate_sequence(unsigned char* seq,struct parameters*  param)
{

	int j,c,errors;
	char tmp_seq[MAX_LINE]; 
	char alphabet[] = "ACGTN";
	float a,b;
	char subc;
	
	b  = ((param->sim_end_error - ((float)param->sim_length *  param->sim_start_error)) / (float)(param->sim_length -1)) * -1.0f;
	
	a = param->sim_start_error - b;
	c = 0;
	errors = 0;
	for(j = 0; j < param->sim_length;j++){
		//	fprintf(stderr,"%d	%f\n",j, (float)(j+1)* a +b);
		if(ran1(seed)  <= (float)(j+1)* a +b){
			errors++;
			if(ran1(seed) <=  param->sim_mismatch_prob){
				subc = alphabet[random_int(4)];
				//while(subc ==  alphabet[seq[j]]){
				//	subc = alphabet[random_int(4)];
				//}
				tmp_seq[c] = subc;
				c++;
				//fprintf(stderr,"MISMATCH\n");
			}else{
				if(ran1(seed) <= 0.5f){
					tmp_seq[c] = alphabet[random_int(4)];
					c++;
					tmp_seq[c] = alphabet[seq[j]];
					c++;
					//	fprintf(stderr,"INSERT\n");
				}else{
					//	fprintf(stderr,"DELETION\n");
				}
			}
		}else{
			tmp_seq[c] = alphabet[seq[j]];
			c++;
		}
	}
	for(j = 0; j < c;j++){
		seq[j] = tmp_seq[j];
	}
	seq[c] = 0;
	return errors;
}





